'-------------------------------------------------------------------------------
' Thermometer Clock
' DigitalTempV7.bas
' Displays time, and temp from upto two external temp. sensors
' (C) Joe Farr, 2012, 2013
'-------------------------------------------------------------------------------
' I/O Assignments:
' A0 - RTC_SCLK
' A1 - RTC_IO
' A2 - RTC_RST
' A3 - DQ2
' A4 - TIME_BUTTON Input
' A5 - DQ1
' A6 - XTAL
' A7 - XTAL
' B0 - Seg: d.p.
' B1 - Seg:
' B2 - Seg:
' B3 - Seg:
' B4 - Seg: g
' B5 - Seg:
' B6 - Seg:
' B7 - Seg:
' C0 - Digit 0
' C1 - Digit 1
' C2 - Digit 2
' C3 - Digit 3
' C4 - TimeLED2
' C5 - TimeLED1
' C6 - SensorLed1
' C7 - SensorLed2
'
'-------------------------------------------------------------------------------
' Version Details
' 1.0    12/11/11     Initial prototype versions
' 5.0    15/05/12     First release
' 6.0    14/04/13     Updated for brighter / faster display refresh & simpler 
'                     code maintenance.
' 7.0    17/08/13     Corrected a problem with the backup battery not charging
'-------------------------------------------------------------------------------

'-------------------------------------------------------------------------------
' Additional info datasheets
' RTC (Real Time Clock)
' http://datasheets.maxim-ic.com/en/ds/DS1302.pdf
' Digital Thermometer 
' http://datasheets.maximintegrated.com/en/ds/DS18B20.pdf

'-------------------------------------------------------------------------------


Device 18F25K22
Xtal 64

Include "Timers.inc" ' Load the Timer macros into the program

' -----------------------------------------------------------------------------------------------------------
' Configure the PICs internal resources                 
' -----------------------------------------------------------------------------------------------------------
    All_Digital True 

   'Use For External 16MHz Xtal Then PLL x4 = 64MHz
   Config_Start
        FOSC = HSMP ; High Speed Medium Power        
        PLLCFG = On ; Oscillator multiplied by 4
        PRICLKEN = On ; Primary clock enabled
        FCMEN = Off ; Fail-Safe Clock Monitor disabled
        PWRTEN = Off ; Power up timer disabled
        BOREN = Off ; Brown-out Reset disabled in hardware and software
        WDTEN = Off ; Watch dog timer is always disabled. SWDTEN has no effect.        
    Config_End
    
' -----------------------------------------------------------------------------------------------------------
' If you don't want to use a 16MHz crystal and/or want to make use of I/O pins PORTA.6 and PORTA.7
' disable the above section for External crystal, and enable this section which allows running from the 
' PIC's internal oscillator. Since the RTC is itself cystal controlled, the clocks accuracy is NOT affected
' -----------------------------------------------------------------------------------------------------------
' Use Internal PIC clock running at 16MHh then PLL x4 = 64MHz (uncomment this section but you MUST comment out the above one)
'    Config_Start
'        FOSC = INTIO67 ; Internal OSC, PORTA.6 and PORTA.7 available as I/O pins
'        FCMEN = Off ; Fail-Safe Clock Monitor disabled
'        PWRTEN = Off ; Power up timer disabled
'        BOREN = Off ; Brown-out Reset disabled in hardware and software
'        WDTEN = Off ; Watch dog timer is always disabled. SWDTEN has no effect.        
'    Config_End
'
'    ' Check datasheet For more details on OSCCON and OSCTUNE
'    ' OSCCON - IRCV<2:0>  
'    ' 111 = HFINTOSC (16 MHz)  
'    OSCCON.6 = 1    ' IRCV<2> 
'    OSCCON.5 = 1    ' IRCV<1>    
'    OSCCON.4 = 1    ' IRCV<0>     
'    OSCTUNE.6 = 1   ' Use PLL for x4 Speed            
' -----------------------------------------------------------------------------------------------------------

' -----------------------------------------------------------------------------------------------------------
' Symbols & Constants
' -----------------------------------------------------------------------------------------------------------
    Symbol dspSegment   = LATB
    Symbol dspDigit     = LATC
    Symbol DQ1          = PORTA.5
    Symbol DQ2          = PORTA.3
    
    Symbol SensorLed1   = LATC.6
    Symbol SensorLed2   = LATC.7
    Symbol TimeLED1     = LATC.5
    Symbol TimeLED2     = LATC.4
    
    Symbol RTC_RST      = PORTA.2
    Symbol RTC_IO       = PORTA.1    
    Symbol RTC_SCLK     = PORTA.0

    Symbol TIME_BUTTON  = PORTA.4    

' -----------------------------------------------------------------------------------------------------------
' Configue I/O pins
' -----------------------------------------------------------------------------------------------------------
    ' Force all ports to output... 
    Output PORTA
    Output PORTB
    Output PORTC           
    '...override As requried
    Input TIME_BUTTON
 
' -----------------------------------------------------------------------------------------------------------
' Variables & Arrays
' -----------------------------------------------------------------------------------------------------------   
    Dim bytDispDecode[12]   As Byte    
    Dim bytRTC[9]           As Byte     ' Year, Day, Month, Date, Hr, Min, Sec, WP, Control
    Dim bytBuffer[4]        As Byte     ' Main display buffer
    Dim bytBufferTmp[4]     As Byte     ' Holds a working copy copy of the display buffer untill ready to be used    
    Dim dwValue             As Dword
    Dim wTempValue          As Word 
    Dim wDisplayDuration    As Word
    Dim wClockSetDelay      As Word
    Dim bytClockSetLoop     As Byte 
    Dim bytTemp             As Byte 
    Dim bytTemp1            As Byte
    Dim bytTemp2            As Byte 
    Dim bytDigitCount       As Byte 
    Dim bytCPerD            As Byte 
    Dim bytBIN2BCD_VAL      As Byte 
    Dim bytBCD2BIN_VAL      As Byte
    Dim bytMins             As Byte 
    Dim bytLow_Bits         As Byte 
    Dim bytHigh_Bits        As Byte 
    Dim bytTempSeconds      As Byte 
    Dim bytTempMinutes      As Byte 
    Dim bytTempSensorNumber As Byte 
    Dim bDisplayingTime     As Bit  
    Dim bRTCPresent         As Bit
    Dim bProgramStarting    As Bit 
    Dim bTempSensor1Present As Bit
    Dim bTempSensor2Present As Bit 
    Dim bFlag               As Bit 
          
    
        
' -----------------------------------------------------------------------------------------------------------
' Initialise Variables
' -----------------------------------------------------------------------------------------------------------  
    
    '********************************************************************
    ' Digit decode table. Controls which segments on a digit need to be
    ' illuminated to display a numeric value
    '********************************************************************
    bytDispDecode[0] = 238 ' 0
    bytDispDecode[1] = 130 ' 1
    bytDispDecode[2] = 220 ' 2
    bytDispDecode[3] = 214 ' 3
    bytDispDecode[4] = 178 ' 4
    bytDispDecode[5] = 118 ' 5
    bytDispDecode[6] = 126 ' 6
    bytDispDecode[7] = 194 ' 7
    bytDispDecode[8] = 254 ' 8
    bytDispDecode[9] = 242 ' 9
    bytDispDecode[10] = 0 ' Blank digit
    bytDispDecode[11] = 16 ' Minus Sign
            
    Clear bytBuffer ' Make sure theres no rubbish in the display buffer - clear to all zeros
    bDisplayingTime = 0
    bytDigitCount = 0
    bTempSensor1Present = 1 ' Assume the external devices are present, till we prove otherwise
    bTempSensor2Present = 1  
    bRTCPresent = 1 

' -----------------------------------------------------------------------------------------------------------
' Initialise Hardware
' -----------------------------------------------------------------------------------------------------------      
    ' Sensor/Time sep 5mm LEDs off
    Low SensorLed1 
    Low SensorLed2 
    Low TimeLED1
    Low TimeLED2    
        
    '***************************************************************
    ' Configure the RTC    
    ' Reset the RTC before we do anything
    Low RTC_RST 
    Low RTC_SCLK 
    ' Read the RTC settings (including the date/time)
    GoSub GetTime
    
    ' Clear the ClockHalt flag - else RTC wont run
    bytRTC#5.7 = 0   ' rtc[5] bit 7        
            
    ' Setup the charge control options    
    bytRTC[7] = 165 ' Set charge rate for NiHM battery - 2.2mA (this is the max. charge rate option)
    GoSub SetBattChargeRate
    
    ' Set to 24hr Mode - (Change To 1 For 12hr Mode)
    bytRTC#3.7 = 0 ' rtc[3] bit 7    
    GoSub SetTime  ' Write the settings back to the RTC            
    '
    ' Now read-back what we've written and check it
    GoSub GetTime
    If bytRTC#3.7 <> 0 Then 
        ' Dosn't look like the RTC is installed :(
        bRTCPresent = 0
    End If        
    
' -----------------------------------------------------------------------------------------------------------
' All initialisation complete
' Specify the main interrupt routine, and start the main routine loop runinng

    On_Hardware_Interrupt GoTo ScanDisplay ' Point to interrupt handler
    GoTo MainProgramStart ' Jump over the interrupt handler and any subroutines
            
' -----------------------------------------------------------------------------------------------------------
' Interrupt handler - Scan the display and output the contents of the 'bytBuffer' array
' -----------------------------------------------------------------------------------------------------------
ScanDisplay:
    Context Save
    If INTCONbits_T0IF = 1 Then   
        ' Interrupt was called because of Timer0 overflow       
        ' Disable all the display Commons
        dspDigit.0 = 0        
        dspDigit.1 = 0
        dspDigit.2 = 0
        dspDigit.3 = 0
        
        ' Extract the value to display on the selected digit
        bytTemp = bytBuffer[ bytDigitCount ]
                
        ' Decide what it is we are displaying
        If bDisplayingTime = 0 Then 
            ' We only do this for temp display
            If bytDigitCount <= 1 Then 
                ' Suppress leading two zeros
                If bytTemp = 0 Then 
                    bytTemp = 10 ' Blank Digit
                End If
            End If
            bytTemp1 = bytDispDecode[ bytTemp ]
            If bytDigitCount = 2 Then 
                Inc bytTemp1 ' Add the DP
            End If
        Else
            ' Displaying time
            bytTemp1 = bytDispDecode[ bytTemp ]            
        End If

               
        ' Output the segment pattern        
        dspSegment = bytTemp1 
        ' Select the digit
        SetBit dspDigit, 3 - bytDigitCount                    
        
        Inc bytDigitCount
        If bytDigitCount > 3 Then   
            bytDigitCount = 0
        End If  
        
        Clear INTCONbits_T0IF ' Clear the Timer0 overflow flag      
    End If
    Context Restore 
' ------------------------------------ END OF INTERRUPT HANDLER ---------------------------------------------    

            
' -----------------------------------------------------------------------------------------------------------
' MAIN START OF PROGRAM START 
' -----------------------------------------------------------------------------------------------------------
MainProgramStart:
    bProgramStarting = 1
    
    bytTempSensorNumber = 1
    GoSub ReadTemp
    bytTempSensorNumber = 2
    GoSub ReadTemp
      
    bProgramStarting = 0    
    
    ' Define a Timer - 0, it's an 8 bit counter that is incremented from the internal PIC clock
    ' once every 64 clock cycles.

    Symbol TIMERMASK = TIMER_INT_ON & T0_8BIT & T0_SOURCE_INT & T0_PS_1_64    
    OpenTimer0(TIMER_INT_ON & TIMERMASK)    
    INTCONbits_GIE = 1 ' Enable global interrupts    

' -----------------------------------------------------------------------------------------------------------
' MAIN START OF PROGRAM LOOP
' -----------------------------------------------------------------------------------------------------------
MainLoop:    
    ' First lets display the time (if there is an RTC installed)
    ' We want this for 30 seconds
    If bRTCPresent = 1 Then 
        wDisplayDuration = 30 ' Seconds
        GoSub ShowTime
    End If

    ' Next, the temp from sensor 1 if that's present
    If bTempSensor1Present = 1 Then 
        wDisplayDuration = 10 ' Seconds    
        GoSub ShowTemp1
    End If

    ' Lastly, the temp from sensor 2 if that's present    
    If bTempSensor2Present = 1 Then 
        wDisplayDuration = 10 ' Seconds   
        GoSub ShowTemp2
    End If

    '
    ' That's it... all tasks completed.
    ' Start all over again
    GoTo MainLoop    
    
' -----------------------------------------------------------------------------------------------------------
' Show the current time 
' Keep this on the display for wDisplayTime seconds
' -----------------------------------------------------------------------------------------------------------
ShowTime:
    bytTempSeconds = 99 ' Force a toggle of the seconds indicators
    bytTempMinutes = 99 ' Force as update to the minutes display.
        
    ' Within this routine, there is a 100ms delay. This allows for the push button to be read fairly quickly.
    ' So, there are 10 x 100ms in 1000ms (1 second), so we multiply the required display
    ' time (wDisplayDuration) by 10.    
    wDisplayDuration = wDisplayDuration * 10
    
    While wDisplayDuration > 0
        GoSub DisplayTime
                
        Dec wDisplayDuration
                
        If TIME_BUTTON = 0 Then 
            ' User has pushed the time setting button
            GoSub IncrementTime 
            wDisplayDuration = 10 ' Default to 10 seconds of time display when returning from time setting routine
            GoTo ShowTime ' Restart the full time display window
        End If
        
        DelayMS 100
    Wend                 
    
    TimeLED1 = 0
    TimeLED2 = 0

    Return    

' -----------------------------------------------------------------------------------------------------------
' Show the current time 
' Keep this on the display for wDisplayTime seconds
' -----------------------------------------------------------------------------------------------------------
DisplayTime:    
    GoSub GetTime ' Read current time

DisplayCurrentBufferedTime:              
    If bytRTC[4] <> bytTempMinutes Then 
        bytTempMinutes = bytRTC[4]
        ' Decode the Hrs 
        bytBCD2BIN_VAL = bytRTC[3]        
        GoSub BCD_TO_BIN    
        bytBufferTmp[0] = Dig bytBCD2BIN_VAL, 1        
        bytBufferTmp[1] = Dig bytBCD2BIN_VAL, 0

        ' Decode the Mins                
        bytBCD2BIN_VAL = bytRTC[4]        
        GoSub BCD_TO_BIN  
        bytBufferTmp[2] = Dig bytBCD2BIN_VAL, 1         
        bytBufferTmp[3] = Dig bytBCD2BIN_VAL, 0
        
        bDisplayingTime = 1 ' Make sure the display scan routine knows we are displaying the time and not Temp         
        GoSub TransferBuffer
    End If
            
    If bytRTC[5] <> bytTempSeconds Then 
        ' The seconds value has change from the last time we read it
        ' so flash (toggle their state) the time seperator LEDs
        bytTempSeconds = bytRTC[5] ' Make a note of the current seconds value for next time
        Toggle TimeLED1
        Toggle TimeLED2    
    End If  
    
    Return
    
' -----------------------------------------------------------------------------------------------------------    
IncrementTime:        
    DelayMS 5 ' Debounce the button    
    bytClockSetLoop = 60
    wClockSetDelay = 500
    bDisplayingTime = 1

    GoSub GetTime ' Read current time
    While TIME_BUTTON = 0 ' Time button still pressed
            
        bytBCD2BIN_VAL = bytRTC[4] ' Mins
        GoSub BCD_TO_BIN          
        Inc bytBCD2BIN_VAL ' Increment the minutes
        bytMins = bytBCD2BIN_VAL 
        If bytMins > 59 Then 
            ' Can't have more than 59 mins in a hour
            bytMins = 0            
            ' Increment hours
            bytBCD2BIN_VAL = bytRTC[3] ' Hrs
            GoSub BCD_TO_BIN                      
            Inc bytBCD2BIN_VAL
            If bytRTC#3.7 = 0 Then 
                ' 24 Hr Mode
                If bytBCD2BIN_VAL > 23 Then 
                    bytBCD2BIN_VAL = 0
                End If
            Else
                ' 12 Hr Mode
                If bytBCD2BIN_VAL > 11 Then 
                    bytBCD2BIN_VAL = 0
                End If
            End If                
            bytBIN2BCD_VAL = bytBCD2BIN_VAL
            GoSub BIN_TO_BCD
            bytRTC[3] = bytBIN2BCD_VAL ' Hrs
        End If
    
        ' Store the minutes (Either the value incremented or Zero)
        bytBIN2BCD_VAL = bytMins
        GoSub BIN_TO_BCD
        bytRTC[4] = bytBIN2BCD_VAL ' Mins            

        GoSub SetTime            
        bytTempMinutes = 99 ' Force as update to the minutes display. 
        GoSub DisplayCurrentBufferedTime   

        ' Set the rate that we increment the time in clock-setting mode
        Select Case bytClockSetLoop
        Case 0   
            wClockSetDelay = 10 ' Maximum scroll speed
            
        Case < 50
            wClockSetDelay = 100 ' Medium scroll speed
            Dec bytClockSetLoop
            
        Case Else
            ' Slow scroll speed
            Dec bytClockSetLoop
            
        End Select 
                            
        DelayMS wClockSetDelay
        
    Wend
    ' The button has been released
    ' Save the time we've set    
    bytRTC[5] = 0 ' Always reset the seconds        
    GoSub SetTime       
            
    Return

' -----------------------------------------------------------------------------------------------------------    
ShowTemp1:
    SensorLed1 = 1 ' Temp sensor LED On
    bytTempSensorNumber = 1
    GoTo ShowTemp ' ShowTemp routine does the executes the Return

' -----------------------------------------------------------------------------------------------------------    
ShowTemp2:
    SensorLed2 = 1 ' Temp sensor LED On
    bytTempSensorNumber = 2
    GoTo ShowTemp ' ShowTemp routine does the executes the Return 

' -----------------------------------------------------------------------------------------------------------    
ShowTemp:
    ' Read the temp just the once else the display can hop around.
    GoSub ReadTemp ' We do this twice as sometimes it dosn't work the first time
    GoSub ReadTemp    
    GoSub TransferBuffer
    bDisplayingTime = 0 ' Buffer no-longer contains time data    
    While wDisplayDuration > 0                
        DelayMS 1000     
        Dec wDisplayDuration
    Wend

    SensorLed1 = 0 ' Temp sensor LEDs Off
    SensorLed2 = 0 
        
    Return
                
    
' -----------------------------------------------------------------------------------------------------------
' Application SubRoutines
' -----------------------------------------------------------------------------------------------------------
'
' -----------------------------------------------------------------------------------------------------------
SetBattChargeRate:
    ' Check the datasheet for what all these mean    
    OpenTimer0(TIMER_INT_OFF & TIMERMASK)  
    RTC_RST = 1 'enable data tx
    SHOut RTC_IO, RTC_SCLK, LsbFirst, [ $8e ,0 ] 'enable write
    RTC_RST = 0 'reset RTC
    
    RTC_RST = 1 'ready for tx
    SHOut RTC_IO, RTC_SCLK, LsbFirst, [ $90, bytRTC[7] ]
    RTC_RST = 0 'reset RTC
    OpenTimer0(TIMER_INT_ON & TIMERMASK)  
    
    Return

' -----------------------------------------------------------------------------------------------------------
SetTime: 'set the time
    ' Check the datasheet for what all these mean    
    OpenTimer0(TIMER_INT_OFF & TIMERMASK)  
    RTC_RST = 1 'enable data tx
    SHOut RTC_IO, RTC_SCLK, LsbFirst, [ $8e ,0 ] 'enable write
    RTC_RST = 0 'reset RTC
    
    RTC_RST = 1 'ready for tx
    'write all 8 RTC registers in burst mode
    bytRTC[6] = 0
    SHOut RTC_IO, RTC_SCLK, LsbFirst, [ $be, bytRTC[5], bytRTC[4], bytRTC[3], bytRTC[2], bytRTC[1], bytRTC[6], bytRTC[0], 0]
    RTC_RST = 0 'reset RTC
    OpenTimer0(TIMER_INT_ON & TIMERMASK)  
    
    Return

' -----------------------------------------------------------------------------------------------------------
GetTime: 'read the time from the RTC
    ' Check the datasheet for what all these mean
    OpenTimer0(TIMER_INT_OFF & TIMERMASK)  
    RTC_RST = 1 'ready for tx
    SHOut RTC_IO, RTC_SCLK, LsbFirst, [ $bf ] 'Read all 8 RTC registers in burst mode
    SHIn RTC_IO, RTC_SCLK, LsbPre, [ bytRTC[5], bytRTC[4], bytRTC[3], bytRTC[2], bytRTC[1], bytRTC[6], bytRTC[0], 0 ]
    RTC_RST = 0 ' Reset RTC
    OpenTimer0(TIMER_INT_ON & TIMERMASK)  

    Return
 
' -----------------------------------------------------------------------------------------------------------
' BINARY to BCD conversion
' BIN2BCD_VAL is the input and output variable
BIN_TO_BCD:
    bytLow_Bits = bytBIN2BCD_VAL // 10          'get lsb,same for Bin and BCD
    bytHigh_Bits = bytBIN2BCD_VAL / 10	      'get msb
    bytBIN2BCD_VAL = bytHigh_Bits * 16	      'covert msb to BCD
    bytBIN2BCD_VAL = bytBIN2BCD_VAL + bytLow_Bits 'add BCD msb and lsb together

    Return
  
' -----------------------------------------------------------------------------------------------------------
' BCD to BINARY conversion
' BCD2BIN_VAL is the input and output variable
BCD_TO_BIN:
    bytTemp1 = bytBCD2BIN_VAL & $F ' Convert the values from BCD to BIN
    bytTemp2 = bytBCD2BIN_VAL & $F0 ' Mast off either side
    bytTemp2 = bytTemp2 >> 4 ' divide by 16
    bytTemp2 = bytTemp2 * 10 ' X by 10
    bytBCD2BIN_VAL = bytTemp1 + bytTemp2 ' Now add the original number
    
    Return     

' -----------------------------------------------------------------------------------------------------------
' Read Temp Sensor #1 or #2
' bytTempSensorNumber = 1 or 2 and is the sensor to read
ReadTemp:
    OpenTimer0(TIMER_INT_OFF & TIMERMASK)  
    Select Case bytTempSensorNumber
    Case 1
        OWrite DQ1, 1, [$CC, $44] ' Send Calculate Temperature command to temp sensor 1
        OWrite DQ1, 1, [$CC, $BE] ' Send Read ScratchPad command    
        ORead DQ1, 2,[wTempValue.LowByte,wTempValue.HighByte, bytTemp, bytTemp, bytTemp, bytTemp, bytTemp, bytCPerD]        

    Case 2
        OWrite DQ2, 1, [$CC, $44] ' Send Calculate Temperature command to temp sensor 2
        OWrite DQ2, 1, [$CC, $BE] ' Send Read ScratchPad command    
        ORead DQ2, 2,[wTempValue.LowByte,wTempValue.HighByte, bytTemp, bytTemp, bytTemp, bytTemp, bytTemp, bytCPerD]        
    End Select    
    OpenTimer0(TIMER_INT_ON & TIMERMASK)  
     
    ' Load the display buffer with the value we've just read from the temp sensor
    ' The "* 625" is to do with the way the sensor works... check the datasheet 
    If GetBit wTempValue, 15 = 1 Then 
        ' Negative
        dwValue = ($10000 - wTempValue) * 625
        bytBufferTmp[0] = 11 ' Display pattern for a Minus sign
    Else
        dwValue = wTempValue * 625
        bytBufferTmp[0] = Dig dwValue, 6    
    End If

    bytBufferTmp[1] = Dig dwValue, 5
    bytBufferTmp[2] = Dig dwValue, 4
    bytBufferTmp[3] = Dig dwValue, 3        
    
    If bProgramStarting  = 1 Then 
        ' This is the initial pass to see which temp sensors are present
        If bytBufferTmp[1] + bytBufferTmp[2] + bytBufferTmp[3] = 0 Then 
            bFlag = 0 ' Not found
        Else
            bFlag = 1 ' Temp sensor found
        End If
        
        Select Case bytTempSensorNumber
        Case 1
            bTempSensor1Present = bFlag
                        
        Case 2
            bTempSensor2Present = bFlag            
        End Select         
    End If

    Return
    
' -----------------------------------------------------------------------------------------------------------
TransferBuffer:
    ' We use bytBufferTmp so that routines can be building up the data in the display buffer
    ' and then it can be transfered to the "real" display buffer all in one go
    bytBuffer[ 0 ] = bytBufferTmp[ 0 ]
    bytBuffer[ 1 ] = bytBufferTmp[ 1 ]
    bytBuffer[ 2 ] = bytBufferTmp[ 2 ]
    bytBuffer[ 3 ] = bytBufferTmp[ 3 ]
    
    Return        
    
    




                
    

